{% extends "base.html" %}

<!-- This is where you can tweak your theme -->
<!-- For more details, see https://squidfunk.github.io/mkdocs-material/customization/#overriding-blocks -->

{% block announce %}{% endblock %}
