"""
The core module defines the core functionalities in doXtrings execution pipeline.
The execution pipeline is:
    search: search for the files to be analyzed
    parse: parse the files found
    check: check the parsed documentables found
    report: report the findings of the analysis
"""
